/*
 * Copyright 2002-2010 Guillaume Cottenceau.
 *
 * This software may be freely redistributed under the terms
 * of the X11 license.
 *
 */

#include <prg4/prgimg.h>

#include <stdlib.h>

#define PNG_DEBUG 3
#include <png.h>

struct prgimg_t {
    png_byte colorModel;
    int width, height;
    png_bytep * row_pointers;
};


png_byte color_type;
png_byte bit_depth;

int number_of_passes;

prgimg* prgimg_create_empty(int width, int height){
    prgimg* img = (prgimg*) malloc(sizeof (prgimg));
    if (!img) {
        fprintf(stderr, "[create_prgimg_empty] malloc failed\n");
        return NULL;
    }

    img->colorModel = PNG_COLOR_TYPE_RGBA;
    img->width = width;
    img->height = height;
    img->row_pointers = (png_bytep*) malloc(sizeof (png_bytep) * img->height);
     if (!img->row_pointers) {
        fprintf(stderr, "[create_prgimg_empty] Error with allocating %ld bytes \n", sizeof (png_bytep) * img->height);
        free(img);
        return NULL;
    }
    for (int y = 0; y < img->height; y++) {
        img->row_pointers[y] = (png_byte*) malloc(4*sizeof(png_byte)*img->width);
        memset(img->row_pointers[y],255,4*sizeof(png_byte)*img->width);
        if (!img->row_pointers[y]) {
            fprintf(stderr, "[create_prgimg_empty] Error with allocating %lu bytes \n", 4*sizeof(png_byte)*img->width);
            for(y--;y>=0; y--){
               free(img->row_pointers[y]);
            }
            free(img->row_pointers);
            free(img);
            return NULL;
        }
    }

    return img;
    
}

prgimg* prgimg_read_file(const char* file_name) {


    unsigned char header[8]; // 8 is the maximum size that can be checked

    /* open file and test for it being a png */
    FILE *fp = fopen(file_name, "rb");
    if (!fp) {
        fprintf(stderr, "[read_prgimg_file] File %s could not be opened for reading\n", file_name);
        return NULL;
    }
    fread(header, 1, 8, fp);
    if (png_sig_cmp(header, 0, 8)) {
        fprintf(stderr, "[read_prgimg_file] File %s is not recognized as a PNG file\n", file_name);
        fclose(fp);
        return NULL;
    }


    prgimg* img = (prgimg*) malloc(sizeof (prgimg));
    if (!img) {
        fprintf(stderr, "[read_prgimg_file] malloc failed\n");
        fclose(fp);
        return NULL;
    }
    /* initialize stuff */
    png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

    if (!png_ptr) {
        fprintf(stderr, "[read_prgimg_file] png_create_read_struct failed\n");
        free(img);
        fclose(fp);
        return NULL;
    }

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) {
        fprintf(stderr, "[read_prgimg_file] png_create_info_struct failed\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        free(img);
        fclose(fp);
        return NULL;
    }

    if (setjmp(png_jmpbuf(png_ptr))) {
        fprintf(stderr, "[read_prgimg_file] Error during init_io\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        free(img);
        fclose(fp);
        return NULL;
    }

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, 8);

    png_read_info(png_ptr, info_ptr);

    img->width = png_get_image_width(png_ptr, info_ptr);
    img->height = png_get_image_height(png_ptr, info_ptr);
    color_type = png_get_color_type(png_ptr, info_ptr);
    bit_depth = png_get_bit_depth(png_ptr, info_ptr);

    number_of_passes = png_set_interlace_handling(png_ptr);
    png_read_update_info(png_ptr, info_ptr);


    /* read file */
    if (setjmp(png_jmpbuf(png_ptr))) {
        fprintf(stderr, "[read_prgimg_file] Error during read_image\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        free(img);
        fclose(fp);
        return NULL;
    }

    img->row_pointers = (png_bytep*) malloc(sizeof (png_bytep) * img->height);
    if (!img->row_pointers) {
        fprintf(stderr, "[read_prgimg_file] Error with allocating %ld bytes \n", sizeof (png_bytep) * img->height);
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        free(img);
        fclose(fp);
        return NULL;
    }
    for (int y = 0; y < img->height; y++) {
        img->row_pointers[y] = (png_byte*) malloc(png_get_rowbytes(png_ptr, info_ptr));
        if (!img->row_pointers[y]) {
            fprintf(stderr, "[read_prgimg_file] Error with allocating %u bytes \n", (unsigned int) png_get_rowbytes(png_ptr, info_ptr));
            for(y--;y>=0; y--){
               free(img->row_pointers[y]); 
            }
            free(img->row_pointers);
            png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
            free(img);
            fclose(fp);
            return NULL;
        }
    }

    png_read_image(png_ptr, img->row_pointers);
    img->colorModel = png_get_color_type(png_ptr, info_ptr);


    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    
    
    fclose(fp);

    return img;

}

void prgimg_write_file(const char* file_name, prgimg* img) {
    /* create file */
    FILE *fp = fopen(file_name, "wb");
    if (!fp) {
        fprintf(stderr, "[write_prgimg_file] File %s could not be opened for writing\n", file_name);
        return;
    }


    /* initialize stuff */
    png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

    if (!png_ptr) {
        fprintf(stderr, "[write_prgimg_file] png_create_write_struct failed\n");
        fclose(fp);
    }

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) {
        fprintf(stderr, "[write_prgimg_file] png_create_info_struct failed\n");
        png_destroy_write_struct(&png_ptr,&info_ptr);
        fclose(fp);
    }

    if (setjmp(png_jmpbuf(png_ptr))) {
        fprintf(stderr, "[write_prgimg_file] Error during init_io\n");
        png_destroy_write_struct(&png_ptr,&info_ptr);
        fclose(fp);
    }

    png_init_io(png_ptr, fp);


    /* write header */
    if (setjmp(png_jmpbuf(png_ptr))) {
        fprintf(stderr, "[write_prgimg_file] Error during writing header\n");
        png_destroy_write_struct(&png_ptr,&info_ptr);
        fclose(fp);
    }

    png_set_IHDR(png_ptr, info_ptr, img->width, img->height,
            bit_depth, color_type, PNG_INTERLACE_NONE,
            PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

    png_write_info(png_ptr, info_ptr);


    /* write bytes */
    if (setjmp(png_jmpbuf(png_ptr))) {
        fprintf(stderr, "[write_prgimg_file] Error during writing bytes\n");
        png_destroy_write_struct(&png_ptr,&info_ptr);
        fclose(fp);
    }

    png_write_image(png_ptr, img->row_pointers);


    /* end write */
    if (setjmp(png_jmpbuf(png_ptr))) {
        fprintf(stderr, "[write_prgimg_file] Error during end of write\n");
        png_destroy_write_struct(&png_ptr,&info_ptr);
        fclose(fp);
    }

    png_write_end(png_ptr, NULL);


    png_destroy_write_struct(&png_ptr,&info_ptr);
    
    fclose(fp);
}

void prgimg_free(prgimg** iimg) {
    prgimg* img = *iimg;
    /* cleanup heap allocation */
    for (int y = 0; y < img->height; y++)
        free(img->row_pointers[y]);
    free(img->row_pointers);

    free(img);
    iimg = NULL;
}

unsigned char _myBlackPixel = 0;

unsigned char* get_prgimg_pixel_color(const prgimg const* img, int x, int y, int offset) {
    if (img->colorModel == PNG_COLOR_TYPE_RGB) {
        fprintf(stderr, "[get_prgimg_pixel_color] input file is PNG_COLOR_TYPE_RGB but must be PNG_COLOR_TYPE_RGBA "
                "(lacks the alpha channel)\n");
        return &_myBlackPixel;
    }

    if (img->colorModel != PNG_COLOR_TYPE_RGBA) {
        fprintf(stderr, "[get_prgimg_pixel_color] color_type of input file must be PNG_COLOR_TYPE_RGBA (%d) (is %d)\n",
                PNG_COLOR_TYPE_RGBA, img->colorModel);
        return &_myBlackPixel;
    }

    png_byte* row = img->row_pointers[y];
    png_byte* ptr = &(row[x * 4+offset]);
    return ptr;

}

unsigned char prgimg_get_red(const prgimg * img, int x, int y) {
    return *get_prgimg_pixel_color(img, x, y, 0);
}

unsigned char prgimg_get_green(const prgimg * img, int x, int y) {
    return *get_prgimg_pixel_color(img, x, y, 1);
}

unsigned char prgimg_get_blue(const prgimg * img, int x, int y) {
    return *get_prgimg_pixel_color(img, x, y, 2);
}

void prgimg_set_red(prgimg* img, int x, int y, unsigned char col) {
    unsigned char* c = get_prgimg_pixel_color(img, x, y, 0);
    *c = col;
}

void prgimg_set_green(prgimg* img, int x, int y, unsigned char col) {
    unsigned char* c = get_prgimg_pixel_color(img, x, y, 1);
    *c = col;
}

void prgimg_set_blue(prgimg* img, int x, int y, unsigned char col) {
    unsigned char* c = get_prgimg_pixel_color(img, x, y, 2);
    *c = col;
}

int prgimg_get_height(const prgimg const* img) {
    return img->height;
}

int prgimg_get_width(const prgimg const* img) {
    return img->width;
}
